from decimal import Decimal

from convocations.aws.base import yielder
from convocations.base.utils import dump_yaml


def yield_logs(group, stream, start_time=None, session=None, client=None):
    """
    yields log lines from the specified group and stream from the specified
    start time onwards
    """
    client = client or session.client('logs')
    kwargs = {
        'logGroupName': group,
        'logStreamName': stream,
        'startFromHead': True,
    }
    if start_time:
        kwargs['startTime'] = start_time + 1
    next_token = 'sesame'
    while next_token:
        last_token = next_token
        response = client.get_log_events(**kwargs)
        if not response['events']:
            break
        yield from response['events']
        next_token = response.get('nextForwardToken')
        if next_token == last_token:
            break
        kwargs['nextToken'] = next_token


def get_bucket_size(session, buckets):
    from datetime import datetime, timedelta
    from pytz import utc
    cw = session.client('cloudwatch')
    end_time = datetime.now(tz=utc)
    start_time = end_time - timedelta(days=2)
    metric_name = 'BucketSizeBytes'
    response = yielder(
        cw, 'list_metrics', session=session,
        Namespace='AWS/S3', MetricName=metric_name)
    relevant_metrics = []
    for metric in response:
        bucket = None
        storage_type = None
        for dimension in metric['Dimensions']:
            if dimension['Name'] == 'BucketName':
                bucket = dimension['Value']
            elif dimension['Name'] == 'StorageType':
                storage_type = dimension['Value'].lower()
        if bucket in buckets:
            relevant_metrics.append([bucket, storage_type, metric])
    response = cw.get_metric_data(
        MetricDataQueries=[{
            'Id': f'bucket_{i}_{storage_type}',
            'MetricStat': {
                'Metric': metric,
                'Period': 24 * 60 * 60,
                'Stat': 'Maximum',
                'Unit': 'Bytes',
            },
            'Label': f'{bucket}',
            'ReturnData': True,
        } for i, (bucket, storage_type, metric) in enumerate(relevant_metrics, 1)],
        StartTime=start_time,
        EndTime=end_time,
    )
    rows = []
    for x in response['MetricDataResults']:
        bucket = x['Label']
        storage_type = x['Id'].split('_')[-1]
        timestamps = x['Timestamps']
        if timestamps:
            t = timestamps[0]
            size = Decimal(x['Values'][0])
            size = size / Decimal(1024) / Decimal(1024) / Decimal(1024)
            size.quantize(Decimal('0.1'))
            rows.append([ bucket, t, size, storage_type ])
    return rows



from raft.tasks import task
from .base import AzureTask
from ..base.utils import get_context_value
from ..base.utils import notice, notice_end
from ..base.utils import print_table


@task(help=dict(perms='read, write, or manage'), iterable=['perms'], klass=AzureTask)
def grant_access_to_site(ctx, site_name, app_name, perms=None, **kwargs):
    """
    grants the app with name app_name read and write permissions to the site_name sharepoint site

    this convocation is useful when you have created an azure application
    registration with api permission Sites.Selected and need to authorize
    the application to connect to a particular sharepoint site.
    """
    from affliction.graph_client import SynchronousGraphClient
    notice('tenant_name')
    perms = perms or [ 'read', 'write' ]
    creds = kwargs['creds']
    tenant_id = get_context_value(ctx, 'azure.tenant_id')
    api = SynchronousGraphClient(tenant_id, creds=creds)
    domains = api.initial_domain()
    initial_domain = domains['id']
    notice_end(initial_domain)
    hostname = initial_domain.replace('onmicrosoft', 'sharepoint')
    app_id = None
    app_name = app_name.lower()
    notice('looking up service principal')
    result = api.search_service_principals(app_name)
    for x in result:
        if x['displayName'].lower() == app_name:
            app_id = x['appId']
            break
    if not app_id:
        notice_end(False)
        return
    notice_end(f'{app_id}')
    notice('site id')
    result = api.get_sharepoint_site(hostname, site_name)
    site_id = result['id']
    notice_end(site_id)
    notice('checking existing permissions')
    data = api.get_sharepoint_site_permissions(site_id)
    app_id_exists = False
    existing_permission_id = None
    existing_perms = []
    for x in data['value']:
        permission_id = x['id']
        data = api.get_sharepoint_site_permission(site_id, permission_id)
        roles = data['roles']
        identities = data['grantedToIdentitiesV2']
        for i in identities:
            app = i.get('application')
            if app and app['id'] == app_id:
                app_id_exists = True
                existing_permission_id = permission_id
                existing_perms = roles
                break
        if roles == perms and app_id_exists:
            notice_end('permission already granted')
            return
    notice_end(existing_permission_id is not None)
    if existing_permission_id:
        notice('updating existing permission')
        updated_perms = set(existing_perms)
        for perm in perms:
            updated_perms.add(perm)
        response = api.update_sharepoint_site_permission(
            site_id,
            existing_permission_id,
            roles=list(updated_perms),
            raw=True)
        notice_end(response.ok)
    else:
        notice('creating new permission')
        response = api.create_sharepoint_site_permission(
            site_id, app_id, app_name, perms, raw=True)
        notice_end(response.ok)


@task(klass=AzureTask)
def revoke_site_access(ctx, site_name, app_name, **kwargs):
    """
    revokes all permissions for the app with name app_name from the site_name sharepoint site

    this convocation is useful when you have created an azure application
    registration with api permission Sites.Selected and need to remove a
    previously-granted authorization to connect to a particular sharepoint site.
    """
    from affliction.graph_client import SynchronousGraphClient
    notice('tenant_name')
    creds = kwargs['creds']
    tenant_id = get_context_value(ctx, 'azure.tenant_id')
    api = SynchronousGraphClient(tenant_id, creds=creds)
    domains = api.initial_domain()
    initial_domain = domains['id']
    notice_end(initial_domain)
    hostname = initial_domain.replace('onmicrosoft', 'sharepoint')
    notice('site id')
    result = api.get_sharepoint_site(hostname, site_name)
    site_id = result['id']
    notice_end(site_id)
    notice('checking existing permissions')
    data = api.get_sharepoint_site_permissions(site_id)
    app_name = app_name.lower()
    existing_perms = []
    for x in data['value']:
        permission_id = x['id']
        data = api.get_sharepoint_site_permission(site_id, permission_id)
        identities = data['grantedToIdentitiesV2']
        for i in identities:
            app = i.get('application')
            if app and app['displayName'].lower() == app_name:
                existing_perms.append(permission_id)
    notice_end(bool(existing_perms))
    for x in existing_perms:
        notice(f'revoking {x}')
        response = api.delete_sharepoint_site_permission(site_id, x, raw=True)
        notice_end(response.ok)


@task(klass=AzureTask)
def list_site_permissions(ctx, site_name, **kwargs):
    """
    lists all app permissions for the site_name sharepoint site
    """
    from affliction.graph_client import SynchronousGraphClient
    notice('tenant_name')
    creds = kwargs['creds']
    tenant_id = get_context_value(ctx, 'azure.tenant_id')
    api = SynchronousGraphClient(tenant_id, creds=creds)
    domains = api.initial_domain()
    initial_domain = domains['id']
    notice_end(initial_domain)
    notice('site id')
    hostname = initial_domain.replace('onmicrosoft', 'sharepoint')
    result = api.get_sharepoint_site(hostname, site_name)
    site_id = result['id']
    notice_end(site_id)
    notice('checking existing permissions')
    data = api.get_sharepoint_site_permissions(site_id)
    header = [ 'permission', 'app_id', 'app_name' ]
    rows = []
    for x in data['value']:
        permission_id = x['id']
        data = api.get_sharepoint_site_permission(site_id, permission_id)
        identities = data['grantedToIdentitiesV2']
        app = identities[0].get('application')
        if app:
            roles = data['roles']
            rows.append([ ', '.join(roles), app['id'], app['displayName'] ])
    notice_end()
    print_table(header, rows)



from raft.tasks import task
from ..base import AwsTask


@task(klass=AwsTask)
def gp2_to_gp3(ctx, dry_run=False, session=None, **kwargs):
    """
    converts all gp2 volumes to gp3
    """
    from ...base.utils import print_table, notice, notice_end
    from ..base import yielder
    notice('describing volumes')
    ec2 = session.client('ec2')
    filters = [{
        'Name': 'volume-type',
        'Values': [ 'gp2' ],
    }]
    fn = yielder(ec2, 'describe_volumes', session, Filters=filters)
    gp2_volumes = list(fn)
    notice_end(f'{len(gp2_volumes)}')
    rows = []
    for x in gp2_volumes:
        rows.append([ x['VolumeId'], x['VolumeType'] ])
    print_table([ 'volume', 'type' ], rows)
    if dry_run:
        return
    for volume in gp2_volumes:
        volume_id = volume['VolumeId']
        notice(f'modifying {volume_id}')
        ec2.modify_volume(VolumeId=volume_id, VolumeType='gp3')
        notice_end()


@task(klass=AwsTask)
def available_volumes(ctx, session=None, **kwargs):
    """
    lists all unattached volumes
    """
    from ...base.utils import print_table, notice, notice_end
    from ..base import yielder, name_tag
    notice('describing volumes')
    ec2 = session.client('ec2')
    filters = [{
        'Name': 'status',
        'Values': [ 'available' ],
    }]
    fn = yielder(ec2, 'describe_volumes', session, Filters=filters)
    gp2_volumes = list(fn)
    notice_end(f'{len(gp2_volumes)}')
    rows = []
    for x in gp2_volumes:
        name = name_tag(x) or ''
        rows.append([ x['VolumeId'], x['VolumeType'], x['Size'], name ])
    print_table([ 'volume', 'type', 'size', 'name', ], rows)



import time
from typing import Optional, Any, Iterable, Mapping
from raft.tasks import Task
from raft.context import Context


try:
    from azure.identity._credentials.shared_cache import SharedTokenCacheCredential
    from azure.identity._credentials.shared_cache import _SharedTokenCacheCredential
except ImportError as import_ex:
    print(import_ex)

    class SharedTokenCacheCredential:
        pass

    class _SharedTokenCacheCredential:
        pass


class CachedCredential(_SharedTokenCacheCredential):
    def _get_auth_client(self, **kwargs: Any):
        from azure.identity._internal.aad_client import AadClient
        client_id = kwargs.pop('client_id', None)
        if client_id:
            return AadClient(client_id=client_id, **kwargs)
        return super()._get_auth_client(**kwargs)  # pylint: disable=no-member

    def _get_accounts_having_matching_refresh_tokens(self, is_cae: bool = False):
        """Returns an iterable of cached accounts which have a matching refresh token.

        :param bool is_cae: whether to look in the CAE cache
        :return: an iterable of cached accounts
        :rtype: Iterable[CacheItem]
        """
        import msal
        all_accounts = self._get_cache_items_for_authority(
            msal.TokenCache.CredentialType.ACCOUNT, is_cae)

        accounts = []
        for account in all_accounts:
            # When the token has no family, msal.net falls back to matching client_id,
            # which won't work for the shared cache because we don't know the IDs of
            # all contributing apps. It should be unnecessary anyway because the
            # apps should all belong to the family.
            if account['realm'] == self._tenant_id:
                accounts.append(account)
        return accounts

    def _get_cached_access_token(
        self, scopes: Iterable[str], account: Mapping[str, str], is_cae: bool = False
    ):
        import msal
        from azure.identity import CredentialUnavailableError
        from azure.core.credentials import AccessToken

        cache = self._cae_cache if is_cae else self._cache
        try:
            cache_entries = cache.find(
                msal.TokenCache.CredentialType.ACCESS_TOKEN,
                target=list(scopes),
                query={"realm": account["realm"]},
            )
            for token in cache_entries:
                expires_on = int(token["expires_on"])
                current_time = int(time.time())
                if expires_on - 300 > current_time:
                    return AccessToken(token["secret"], expires_on)
        except Exception as ex:  # pylint:disable=broad-except
            message = f"Error accessing cached data: {ex}"
            raise CredentialUnavailableError(message=message) from ex

        return None

    def _get_account(
        self, username: Optional[str] = None, tenant_id: Optional[str] = None,
        is_cae: bool = False
    ):
        """Returns exactly one account which has a refresh token and matches
        username and/or tenant_id.

        :param str username: an account's username
        :param str tenant_id: an account's tenant ID
        :param bool is_cae: whether to use the CAE cache
        :return: an account
        :rtype: CacheItem
        """
        from azure.identity import CredentialUnavailableError
        accounts = self._get_accounts_having_matching_refresh_tokens(is_cae)
        if not accounts:
            # cache is empty or contains no refresh token -> user needs to sign in
            raise CredentialUnavailableError(
                message=f'no accounts matching {tenant_id}')

        if username:
            filtered_accounts = [ x for x in accounts if x.get('username') == self._username ]
        else:
            filtered_accounts = accounts
        if len(filtered_accounts) == 1:
            return filtered_accounts[0]
        return super()._get_account(username, tenant_id, is_cae)  # pylint: disable=no-member

    def get_token(
        self, *scopes: str, claims: Optional[str] = None, tenant_id: Optional[str] = None, **kwargs: Any
    ):
        from azure.identity import CredentialUnavailableError
        if not scopes:
            raise ValueError("'get_token' requires at least one scope")

        if not self._client_initialized:
            self._initialize_client()

        is_cae = bool(kwargs.get("enable_cae", False))
        token_cache = self._cae_cache if is_cae else self._cache

        # Try to load the cache if it is None.
        if not token_cache:
            token_cache = self._initialize_cache(is_cae=is_cae)

            # If the cache is still None, raise an error.
            if not token_cache:
                raise CredentialUnavailableError(message="Shared token cache unavailable")

        account = self._get_account(self._username, self._tenant_id, is_cae=is_cae)

        token = self._get_cached_access_token(scopes, account, is_cae=is_cae)
        if token:
            return token

        # try each refresh token, returning the first access token acquired
        for refresh_token in self._get_refresh_tokens(account, is_cae=is_cae):
            token = self._client.obtain_token_by_refresh_token(
                scopes, refresh_token, claims=claims,
                tenant_id=tenant_id or self._tenant_id, **kwargs
            )
            return token

        from azure.identity._internal.shared_token_cache import NO_TOKEN
        raise CredentialUnavailableError(message=NO_TOKEN.format(account.get("username")))


class ConvocationsCredential(SharedTokenCacheCredential):
    """Authenticates using tokens in the local cache shared between
    Microsoft applications.

    :param str username: Username (typically an email address) of the user
        to authenticate as. This is used when the
        local cache contains tokens for multiple identities.

    :keyword str authority: Authority of an Azure Active Directory
        endpoint, for example 'login.microsoftonline.com', the authority
        for Azure Public Cloud (which is the default).
        :class:`~azure.identity.AzureAuthorityHosts`
        defines authorities for other clouds.
    :keyword str tenant_id: an Azure Active Directory tenant ID. Used to
        select an account when the cache contains tokens for
        multiple identities.
    :keyword AuthenticationRecord authentication_record: an authentication
        record returned by a user credential such as
        :class:`DeviceCodeCredential` or
        :class:`InteractiveBrowserCredential`
    :keyword cache_persistence_options: configuration for persistent token
        caching. If not provided, the credential
        will use the persistent cache shared by Microsoft development
        applications
    :type cache_persistence_options: ~azure.identity.TokenCachePersistenceOptions
    """

    def __init__(self, username: Optional[str] = None, **kwargs: Any) -> None:
        super().__init__(username, **kwargs)
        self._credential = CachedCredential(
            username=username, **kwargs)


def get_token_info(creds, verbose=True):
    """
    prints the username from the jwt we get from the credentials
    """
    from ..base.utils import notice, notice_end
    # from ..base.utils import dump_yaml
    import jwt
    if verbose:
        notice('logged in as')
    token = creds.get_token('https://management.azure.com/.default')
    data = jwt.decode(token.token.encode('utf8'), options={ 'verify_signature': False })
    # dump_yaml(data, quiet=False)
    try:
        upn = data['unique_name']
    except KeyError:
        upn = f'app/{data["appid"]}'
    if verbose:
        notice_end(upn)
        notice('tid')
    tid = data['tid']
    if verbose:
        notice_end(tid)
        notice('issuer')
    issuer = data['iss']
    if verbose:
        notice_end(issuer)
    if issuer.endswith('/'):
        issuer = issuer[:-1]
    issuer_id = issuer.rsplit('/', 1)[-1]
    tenant_id = issuer_id
    return upn, tenant_id


class AzureTask(Task):
    def __call__(self, *args, **kwargs):
        from azure.identity import ClientSecretCredential
        from azure.identity import ChainedTokenCredential
        from azure.identity import InteractiveBrowserCredential
        from azure.identity import TokenCachePersistenceOptions
        from ..base.utils import get_context_value
        from ..base.utils import notice, notice_end
        ctx = args[0]
        has_context = isinstance(ctx, Context)
        client_id = kwargs.get('client_id')
        client_secret = kwargs.get('client_secret')
        tenant_id = kwargs.get('tenant_id')
        allowed_tenants = kwargs.get('allowed_tenants')
        if allowed_tenants:
            allowed_tenants = allowed_tenants.split(',')
        allow_unencrypted_storage = kwargs.get('allow_unencrypted_storage')
        creds = kwargs.get('creds')
        quiet = kwargs.get('quiet') or False
        cache_name = kwargs.get('cache_name') or None
        verbose = not quiet
        if has_context:
            client_id = client_id or get_context_value(ctx, 'azure.client_id')
            client_secret = client_secret or get_context_value(ctx, 'azure.client_secret')
            tenant_id = tenant_id or get_context_value(ctx, 'azure.tenant_id')
            allow_unencrypted_storage = allow_unencrypted_storage or get_context_value(
                ctx, 'azure.allow_unencrypted_storage')
            cache_name = cache_name or get_context_value(ctx, 'azure.cache_name')
        cache_name = cache_name or 'convocations'
        if verbose:
            notice('client_id')
            notice_end(client_id)
            notice('tenant_id')
            notice_end(tenant_id)
            notice('cache name')
            notice_end(cache_name)
        if not creds and client_id and client_secret and tenant_id:
            creds = ClientSecretCredential(tenant_id, client_id, client_secret)
        elif not creds:
            if verbose:
                notice('using interactive credential')
            options = TokenCachePersistenceOptions(
                name=cache_name,
                allow_unencrypted_storage=allow_unencrypted_storage)
            creds = ChainedTokenCredential(
                ConvocationsCredential(
                    tenant_id=tenant_id,
                    client_id=client_id,
                    cache_persistence_options=options,
                    additionally_allowed_tenants=['*']),
                InteractiveBrowserCredential(
                    tenant_id=tenant_id,
                    client_id=client_id,
                    cache_persistence_options=options,
                )
            )
            if verbose:
                notice_end()
        upn, tenant_id = get_token_info(creds, verbose)
        kwargs['creds'] = creds
        kwargs.setdefault('upn', upn)
        kwargs.setdefault('tenant_id', tenant_id)
        return super().__call__(*args, **kwargs)



from raft.collection import Collection
from .service_tasks import watch_service_deploy
from .task_tasks import list_tasks, task_logs, get_latest_revision



ecs_tasks = Collection()
ecs_tasks.add_task(task_logs)
ecs_tasks.add_task(get_latest_revision)
ecs_tasks.add_task(watch_service_deploy)
ecs_tasks.add_task(list_tasks)



from raft.tasks import task
from ...base.utils import notice, notice_end, print_table
from ..base import AwsTask


@task(klass=AwsTask)
def workspaces(ctx, name='', session=None, **kwargs):
    """
    lists all available workspaces
    """
    from ..base import yielder
    notice('getting workspaces')
    client = session.client('workspaces', region_name='us-east-1')
    rg = yielder(client, 'describe_workspaces', session)
    workspace_map = { x['WorkspaceId']: x for x in rg }
    notice_end(len(workspace_map))
    notice('getting statuses')
    w = list(workspace_map.keys())
    key = 'LastKnownUserConnectionTimestamp'
    while w:
        workspace_ids = w[:25]
        w = w[25:]
        statuses = yielder(
            client,
            'describe_workspaces_connection_status',
            session, WorkspaceIds=workspace_ids)
        statuses = list(statuses)
        for x in statuses:
            workspace_id = x['WorkspaceId']
            last_connection = x.get(key)
            if last_connection:
                last_connection = last_connection.strftime('%Y-%m-%d')
            else:
                last_connection = ''
            workspace_map[workspace_id]['last'] = last_connection
    rows = []
    header = [ 'id', 'user', 'last used' ]
    name = name.lower()
    for x in workspace_map.values():
        username = x['UserName']
        if name in username.lower():
            row = [ x['WorkspaceId'], username, x.get('last', '') ]
            rows.append(row)
    notice_end()
    rows.sort(key=lambda lx: lx[-1])
    rows.reverse()
    print_table(header, rows)



import os
import re
from collections import defaultdict
from datetime import datetime
from datetime import timedelta
from tempfile import NamedTemporaryFile
import time
import boto3
from raft.tasks import task
from .ec2 import instance_by_name, aws_filter
from .base import AwsTask, get_account_id, get_tag
from ..base.utils import notice, notice_end, confirm
from ..base.utils import print_table


def get_latest_ami(name, session):
    ec2 = session.client('ec2')
    responses = ec2.describe_images(Filters=[{
        'Name': 'name',
        'Values': [ f'*{name}*', ],
    }], Owners=[
        session.account_id,
    ])
    images = responses['Images']
    dt = datetime.utcnow()
    latest_image = None
    for image in images:
        creation_date = image['CreationDate']
        creation_date = datetime.strptime(creation_date[:-5], '%Y-%m-%dT%H:%M:%S')
        if dt - creation_date < timedelta(minutes=10):
            latest_image = image
    return latest_image


def load_image(image, session):
    ec2 = session.resource('ec2')
    image_id = image['ImageId']
    return ec2.Image(image_id)


def public_ip(ami_name, session=None):
    """
    Grabs the public IP address from the output in
    cloudformation
    """
    session = session or boto3.DEFAULT_SESSION
    cfn = session.resource('cloudformation')
    notice('getting ip')
    stack = cfn.Stack(f'{ami_name}')
    stack.load()
    outputs = stack.outputs or []
    result = None
    for x in outputs:
        if x['OutputKey'] == 'Ip':
            result = x['OutputValue']
            break
    notice_end(result or 'not found')
    return result


def tail_output_log(ctx, ip, name, session, delay=None):
    """
    In our standard ami builds, the
    user data log file will be written to
    /var/log/{ami_name}-output.log.  Tail it here
    """
    wait_time = 30 if delay is None else int(delay)
    instance = instance_by_name(ctx, name, session=session)
    username = get_tag(instance, 'ssh_user') or 'ubuntu'
    if 'centos' in name:
        username = 'centos'
        wait_time = 90 if delay is None else delay
    elif 'oracle' in name or 'ol7' in name or 'ol8' in name:
        username = 'ec2-user'
        wait_time = 90 if delay is None else delay
    elif 'amazon' in name or 'amzn' in name:
        username = 'ec2-user'
        wait_time = 20 if delay is None else delay
    key_pair = get_tag(instance, 'key_name') or instance.key_name
    if not key_pair.startswith('/'):
        key_pair = f'/{key_pair}'
    notice(f'loading {key_pair} from ssm')
    ssm = session.client('ssm')
    x = ''
    local_key = None
    try:
        try:
            key = ssm.get_parameter(Name=key_pair, WithDecryption=True)
            with NamedTemporaryFile('w', suffix='.key', delete=False) as f:
                f.write(key['Parameter']['Value'])
                local_key = f.name
            os.chmod(local_key, 0o600)
            x = f'-i {local_key}'
            notice_end()
        except Exception as ex:
            notice_end(f'{ex}')
        notice(f'waiting {wait_time} seconds for ssh')
        time.sleep(wait_time)
        notice_end()

        ctx.run(
            f'ssh {x} -o "ServerAliveInterval 2" {username}@{ip}'
            f' "sudo tail -F /var/log/{name}-output.log -F /var/log/cfn-init-cmd.log | '
            f' stdbuf -o0 sed \'/^Cloud-init.*finished at.*/ q\'"', warn=True)
        ctx.run(
            f'ssh {x} -o "ServerAliveInterval 2" {username}@{ip}'
            f' "sudo tail -f /var/log/cfn-init-cmd.log"', warn=True)
        notice_end('----- end of output log -----')
    finally:
        if os.path.exists(local_key):
            os.remove(local_key)


@task(klass=AwsTask)
def wait_for_ami(ctx, name, session=None, delay=None, **kwargs):
    """
    waits for an ami with `name` in it
    """
    # first check if we have a public ip available
    if not name.endswith('-ami'):
        name = f'{name}-ami'
    ip = public_ip(name, session)
    if ip:
        tail_output_log(ctx, ip, name, session, delay=delay)
    name = name[:-4]
    p_image = None
    for i in range(1, 41):
        if p_image:
            notice(f'checking {p_image.id}/{i}')
            p_image.load()
            notice_end(p_image.state)
            if p_image.state == 'available':
                break
        else:
            notice(f'check #{i}')
            latest_image = get_latest_ami(name, session)
            if not latest_image:
                notice_end('not yet initiated')
            else:
                p_image = load_image(latest_image, session)
                notice_end(f'{p_image.id}/{p_image.state}')
                if p_image.state == 'available':
                    break
        time.sleep(15)


@task(positional=['name'], klass=AwsTask, help={
    'name': 'the name of the ami to look for, we will search case-insensitive '
            'for partial match',
    'owners': 'a comma delimited list of owner ami ids to filter by',
    'product_codes': 'a comma delimited list of product codes to filter by',
})
def amis(ctx, name=None, owners=None,
         product_codes=None, quick=False,
         session=None, profile=None, verbose=False, **kwargs):
    """
    Lists all of our amis in aws
    """
    notice('getting images')
    images = []
    if name:
        name = name.lower()
    ec2 = session.client('ec2')
    owners = owners.split(',') if owners else [ 'self' ]
    filters = None
    if product_codes:
        owners.append('aws-marketplace')
        product_codes = product_codes.split(',')
        filters = [ aws_filter('product-code', x) for x in product_codes ]
    params = dict(Owners=owners, Filters=filters)
    if not filters:
        del params['Filters']
    result = ec2.describe_images(**params)
    for x in result['Images']:
        if not name or name in x['Name'].lower():
            images.append(x)
    images.sort(key=lambda lx: lx['CreationDate'])
    images = images[::-1]
    rows = []
    for x in images:
        row = [
            x['ImageId'], x['Name'],
            x['CreationDate'].split('T', 1)[0], x['State'],
        ]
        snapshots = []
        for device_mapping in x['BlockDeviceMappings']:
            if 'Ebs' in device_mapping:
                snapshot_id = device_mapping.get('Ebs', {}).get('SnapshotId')
                snapshot_id = snapshot_id or ''
                snapshots.append(snapshot_id)
        row.append(', '.join(snapshots))
        if verbose:
            row.append(x.get('Description') or '')
        rows.append(row)

    notice_end()
    if quick:
        values = ','.join(x[0] for x in rows)
        print(values)
    else:
        header = ['id', 'name', 'at', 'status', 'snapshot_id' ]
        if verbose:
            header.append('desc')
        print_table(header, rows)


@task(klass=AwsTask, help={
    'image_ids': 'the ami id of the image to delete',
})
def delete_ami(ctx, image_ids, session=None, **kwargs):
    """
    Deletes the ami and any associated snapshots
    """
    ec2 = session.resource('ec2')
    for image_id in image_ids.split(','):
        notice('loading image')
        image = ec2.Image(image_id)
        image.load()
        if not image.meta.data:
            notice_end('not found')
            return
        notice_end(image.name)
        snapshot_ids = []
        snapshots = []
        for x in image.block_device_mappings:
            ebs = x.get('Ebs') or {}
            snapshot_id = ebs.get('SnapshotId')
            if snapshot_id:
                notice(f'loading {snapshot_id}')
                snapshot_ids.append(snapshot_id)
                snapshot = ec2.Snapshot(snapshot_id)
                snapshots.append(snapshot)
                notice_end()
        if not snapshots:
            notice('finding snapshots')
            notice_end('n/a')
        r = confirm(f'You are deleting ami [{image.name}]')
        if not r:
            return

        notice('deregistering ami')
        image.deregister()
        notice_end()
        for snapshot_id, snapshot in zip(snapshot_ids, snapshots):
            notice(f'deleting {snapshot_id}')
            snapshot.delete()
            notice_end()


@task(klass=AwsTask, help={
    'image_id': 'the ami id of the image to delete',
})
def delete_ami_snapshots(ctx, image_id, session=None, **kwargs):
    """
    Deletes the ami and any associated snapshots
    """
    from convocations.aws.base import yielder
    account_id = get_account_id(session)
    ec2 = session.client('ec2')
    notice('finding snapshots')
    snapshots = yielder(ec2, 'describe_snapshots', OwnerIds=[ account_id ])
    rows = []
    for x in snapshots:
        description = x.get('Description') or ''
        snapshot_id = x.get('SnapshotId')
        if image_id in description:
            row = [ snapshot_id, description ]
            rows.append(row)
    notice_end(f'{len(rows)}')
    print_table([ 'id', 'description' ], rows)
    for x, _ in rows:
        r = confirm(f'deleting {x}')
        if r:
            ec2.delete_snapshot(SnapshotId=x)


@task(klass=AwsTask)
def orphaned_ami_snapshots(ctx, delete=False, session=None, **kwargs):
    """
    Finds any ami related snapshots where the original ami has been
    deregistered
    """
    from convocations.aws.base import yielder
    account_id = get_account_id(session)
    ec2 = session.client('ec2')
    notice('finding snapshots')
    snapshots = yielder(ec2, 'describe_snapshots', OwnerIds=[ account_id ])
    snapshots = list(snapshots)
    matcher = re.compile(
        r'^Created by CreateImage\(i\-[0-9a-f]+\) for (?P<image_id>ami\-[0-9a-f]+)$',
    )
    mp_snapshots = defaultdict(list)
    image_ids = set()
    for x in snapshots:
        description = x.get('Description') or ''
        result = matcher.match(description)
        if result:
            image_id = result.group('image_id')
            image_ids.add(image_id)
            x['ImageId'] = image_id
            mp_snapshots[image_id].append(x)
    notice_end(f'{len(snapshots)}')
    notice('loading amis')
    images = yielder(ec2, 'describe_images', Owners=[ account_id ])
    available_image_ids = { x['ImageId'] for x in images }
    notice_end(f'{len(available_image_ids)}')
    unavailable_image_ids = image_ids - available_image_ids
    rows = []
    for x in snapshots:
        image_id = x.get('ImageId')
        if not image_id:
            continue
        description = x.get('Description') or ''
        snapshot_id = x.get('SnapshotId')
        if image_id in unavailable_image_ids:
            row = [ snapshot_id, description ]
            rows.append(row)
    print_table([ 'id', 'description' ], rows)
    if delete:
        for x, _ in rows:
            r = confirm(f'delete {x}')
            if r:
                ec2.delete_snapshot(SnapshotId=x)


@task(klass=AwsTask)
def centos_amis(ctx, session=None, prefix=None, **kwargs):
    """
    finds all marketplace amis released under the centos product code
    """
    amis(ctx, product_codes='aw0evgkw8e5c1q413zgy5pjce', session=session)


@task(klass=AwsTask)
def oracle_amis(ctx, session=None, prefix=None, **kwargs):
    """
    finds all marketplace amis released under the oracle owner id
    """
    amis(ctx, owners='131827586825', session=session, prefix=prefix)


@task(klass=AwsTask)
def ubuntu_amis(ctx, session=None, prefix=None, **kwargs):
    """
    finds all marketplace amis released under the ubuntu owner id
    """
    amis(ctx, owners='099720109477', session=session, prefix=prefix)




import posixpath
import time

from raft.tasks import task
from convocations.aws.base import AwsTask
from convocations.base.utils import print_table, notice, notice_end


def get_task_arns(session, cluster, service, client=None):
    from ..base import yielder
    client = client or session.client('ecs')
    kwargs = {
        'cluster': cluster,
        'serviceName': service,
    }
    tasks = yielder(client, 'list_tasks', **kwargs)
    return list(tasks)


def get_tasks(session, cluster, service, client=None):
    client = client or session.client('ecs')
    arns = get_task_arns(session, cluster, service, client)
    response = client.describe_tasks(
        cluster=cluster,
        tasks=arns,
    )
    return response['tasks']


@task(klass=AwsTask, help={
    'cluster': 'the name of the cluster',
    'service': 'the name of the service',
})
def list_tasks(ctx, cluster, service, session=None, **kwargs):
    """
    lists tasks for a given ecs service in a specified cluster.

    will show a table with the following values;
      * the task id
      * the status of the task (e.g., RUNNING, PENDING, STOPPED)
      * the name of the container as specified in the task definition
      * the image being used by the container (in short form)
    """
    ecs = session.client('ecs')
    notice('getting tasks')
    tasks = get_tasks(session, cluster, service, ecs)
    notice_end(f'{len(tasks)}')
    if not tasks:
        return
    header = [ 'id', 'status', 'name', 'image', ]
    rows = []
    for x in tasks:
        containers = x['containers']
        for container in containers:
            row = [
                container['taskArn'].rsplit('/', 1)[-1],
                container['lastStatus'].lower(),
                container['name'],
                container['image'].split('/', 1)[-1],
            ]
            rows.append(row)
    print_table(header, rows)


def ecs_log_stream(cluster_name: str, task_id: str, session, container_name: str = None):
    """
    Returns (log_group, log_stream, aws_region) for the given ECS task.
    Assumes the task definition is using the awslogs driver.
    """
    ecs = session.client('ecs')

    # 1) Describe the running (or stopped) task to get its taskDefinitionArn
    resp = ecs.describe_tasks(cluster=cluster_name, tasks=[task_id])
    tasks = resp.get('tasks', [])
    if not tasks:
        raise ValueError(
            f'No task found with ID {task_id} in cluster {cluster_name}')
    t = tasks[0]
    task_def_arn = t['taskDefinitionArn']

    # 2) Describe that task definition to read the log config from the container
    task_def = ecs.describe_task_definition(taskDefinition=task_def_arn)['taskDefinition']
    # Assuming a single container definition; pick the one you need if more
    container_definitions = task_def['containerDefinitions']
    container_def = container_definitions[0]
    if container_name:
        container_name = container_name.lower()
        for x in container_definitions:
            if x['name'].lower() == container_name:
                container_def = x
                break
    log_cfg = container_def.get('logConfiguration', {})
    if log_cfg.get('logDriver') != 'awslogs':
        raise RuntimeError('Task is not using the awslogs driver')

    opts = log_cfg['options']
    log_group = opts['awslogs-group']
    stream_pref = opts.get('awslogs-stream-prefix')
    log_region = opts.get('awslogs-region', ecs.meta.region_name)

    # 3) Build the exact stream name
    # ECS uses: {prefix}/{containerName}/{ecsTaskId}
    ecs_task_short_id = t['taskArn'].split('/')[-1]
    pieces = []
    if stream_pref:
        pieces.append(stream_pref)
    pieces.append(container_def['name'])
    pieces.append(ecs_task_short_id)
    stream_name = posixpath.join(*pieces)

    return log_group, stream_name, log_region


def task_status(ctx, cluster_name, task_id, session=None, **kwargs):
    """
    gets the status of a task
    :param ctx:
    :param cluster_name: the name of the cluster
    :param task_id: the id of the task
    :param session: the session passed in from `AwsTask`
    :param kwargs:
    :return:
    """
    ecs = session.client('ecs')
    response = ecs.describe_tasks(
        cluster=cluster_name, tasks=[task_id]
    )
    response = response['tasks'][0]
    status = response['lastStatus']
    if status == 'STOPPED':
        code = response.get('stopCode')
        reason = response.get('stoppedReason')
        if code:
            notice('stop code')
            notice_end(code)
        if reason:
            print(f'{reason}')
        containers = response.get('containers') or []
        for x in containers:
            reason = x.get('reason')
            if reason:
                notice(f'container {x["name"]} stopped with reason')
                notice_end()
                print(reason)
    return status


def print_logs(group_name=None, stream_name=None, start_time=None, client=None, session=None):
    from ..cloudwatch import yield_logs
    lines = list(
        yield_logs(group_name, stream_name, start_time, client=client, session=session)
    )
    lines.sort(key=lambda lx: lx['timestamp'])
    for x in lines:
        print(x['message'])
    if lines:
        start_time = lines[-1]['timestamp']
    return start_time


@task(klass=AwsTask)
def task_logs(ctx, cluster_name, task_id, container_name=None, session=None, **kwargs):
    logs = session.client('logs')
    start_time = None
    notice(f'getting logs for {task_id}')
    group_name, stream_name, _ = ecs_log_stream(cluster_name, task_id, session, container_name)
    notice_end(f'{group_name}/{stream_name}')
    try:
        for n in range(0, 600):
            try:
                start_time = print_logs(group_name, stream_name, start_time, logs, session)
                if n % 5 == 0:
                    status = task_status(ctx, cluster_name, task_id, session=session)
                    if status == 'STOPPED':
                        break
                time.sleep(1)
            except logs.exceptions.ResourceNotFoundException:
                status = task_status(ctx, cluster_name, task_id, session=session)
                if status == 'STOPPED':
                    break
                time.sleep(1)
    except KeyboardInterrupt:
        pass
    time.sleep(3)
    print_logs(group_name, stream_name, start_time, logs, session)


@task(klass=AwsTask)
def get_latest_revision(ctx, task_name, session=None, quiet=False, **kwargs):
    """
    prints out the latest task revision for a task
    :param ctx:
    :param task_name: the name of the task
    :param session: the session passed in from `AwsTask`
    :param quiet: if this flag is set, output is repressed
    :param kwargs:
    :return:
    """
    talkative = not quiet
    if talkative:
        notice('getting task definition')
    ecs = session.client('ecs')
    result = ecs.describe_task_definition(taskDefinition=task_name)
    x = result['taskDefinition']
    result = f"{task_name}:{x['revision']}"
    if talkative:
        notice_end(result)
    return result



import re
from raft.tasks import task
from .base import AzureTask
from ..base.utils import notice, notice_end, print_table


def print_groups(result):
    header = [ 'id', 'name', 'on-prem' ]
    rows = []
    for x in result:
        rows.append([
            x['id'],
            x['displayName'],
            x['onPremisesSyncEnabled'] or '',
        ])
    print_table(header, rows)


def is_uuid(st):
    regex = r'^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$'
    uuid_regex = re.compile(regex, re.IGNORECASE)
    return bool(uuid_regex.match(st))


def resolve_group_id(api, st):
    if is_uuid(st):
        return st
    notice('getting group')
    result = api.search_groups(st)
    if len(result) > 1:
        notice_end('\N{warning sign} \N{warning sign} \N{warning sign} '
                   'multiple groups found')
        print_groups(result)
        return None
    group_id = result[0]['id']
    notice_end(group_id)
    return group_id


def print_members(members):
    header = [ 'id', 'member' ]
    rows = []
    for i, member in enumerate(members, 1):
        rows.append([
            member['id'],
            member['userPrincipalName'].lower(), ])
    rows.sort(key=lambda lx: lx[1])
    print_table(header, rows)
    print()


@task(klass=AzureTask)
def list_group_members(ctx, group_name_or_id, quiet=False, creds=None, tenant_id=None, **kwargs):
    """
    lists all members of a group
    """
    from affliction.graph_client import SynchronousGraphClient
    api = SynchronousGraphClient(tenant_id, creds=creds)
    if is_uuid(group_name_or_id):
        st = group_name_or_id
        members = api.get_group_members(st)
        if not quiet:
            print_members(members)
        else:
            return members
    result = api.search_groups(group_name_or_id)
    all_members = []
    for x in result:
        group_id = x['id']
        name = x['displayName']
        if not quiet:
            notice(f'{name} / {group_id}')
        members = api.get_group_members(group_id)
        all_members += members
        if not quiet:
            notice_end()
            print_members(members)
    if quiet:
        return all_members
    return None


@task(klass=AzureTask)
def groups(ctx, name, creds=None, quiet=False, members=False, tenant_id=None, **kwargs):
    """
    shows all groups with a substring of name
    """
    from affliction.graph_client import SynchronousGraphClient
    api = SynchronousGraphClient(tenant_id, creds=creds)
    name = (name or '').lower()
    params = None
    result = api.search_groups(name, params=params)
    if quiet:
        return result

    print_groups(result)

    header = [ 'member id', 'group', 'member' ]
    if members:
        for x in result:
            group_id = x['id']
            name = x['displayName']
            notice(name)
            members = api.get_group_members(group_id)
            notice_end()
            rows = []
            for i, member in enumerate(members, 1):
                rows.append([
                    member['id'],
                    x['displayName'],
                    member['userPrincipalName'].lower(), ])
            rows.sort(key=lambda lx: lx[2])
            print_table(header, rows)
            print()
    return None


@task(klass=AzureTask, help={
    'name': 'the name of the group to create',
    'description': 'the group description',
    'creds': 'for internal use only',
    'tenant_id': 'for internal use only',
})
def create_group(
        ctx, name, description=None, creds=None, tenant_id=None, **kwargs):
    """
    creates an entra group with the provided name and description
    """
    from affliction.graph_client import SynchronousGraphClient
    api = SynchronousGraphClient(tenant_id, creds=creds)
    notice('creating group')
    result = api.create_group(name, description=description)
    notice_end(result['id'])


@task(klass=AzureTask)
def add_group_members(
        ctx, group_name_or_id, upn, creds=None, tenant_id=None, **kwargs):
    """
    adds members to a group.  users must be specified by upn.  for group,
    group name or id may be given.  if a group name is provided, and more than
    one group matches the name, nothing will be done.
    """
    from affliction.graph_client import SynchronousGraphClient
    api = SynchronousGraphClient(tenant_id, creds=creds)
    # b3e90e13-9a22-46e0-a7a2-739c61eafd80
    group_id = resolve_group_id(api, group_name_or_id)
    if not group_id:
        return
    upns = upn.split(',')
    for x in upns:
        notice(x)
        try:
            user = api.get_user(x)
        except:
            notice_end(False)
            continue
        notice_end(user['id'])
        notice('adding to group')
        api.add_group_member(group_id, user['id'])
        notice_end()

    members = api.get_group_members(group_id)
    print_members(members)
